/* application.c
 *
 * Copyright 2023 Red Hat, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "config.h"

#include "application.h"
#include "window.h"

struct _LunchboxApplication
{
  GtkApplication    parent_instance;
  gboolean          is_decorated;
  gboolean          emulate_touch;
  gboolean          show_host_cursor;
  gboolean          auto_resize;
  gboolean          is_fullscreened;

  /* Subprocess state */
  guint             watcher_id;
  gboolean          process_started;
  GSubprocess      *proc;
  gchar           **argv;
};

G_DEFINE_TYPE (LunchboxApplication, lunchbox_application, GTK_TYPE_APPLICATION)

static void
on_name_appeared_cb (GDBusConnection  *connection,
                     const gchar      *name,
                     const gchar      *name_owner,
                     gpointer          user_data)
{
  LunchboxApplication *self = user_data;
  GtkWindow *window;
  g_assert (LUNCHBOX_IS_APPLICATION (self));
  g_assert (G_IS_DBUS_CONNECTION (connection));
  window = gtk_application_get_active_window (GTK_APPLICATION (self));

  self->process_started = TRUE;
  lunchbox_window_set_display_view (LUNCHBOX_WINDOW (window), connection);
}

static void
on_name_vanished_cb (GDBusConnection  *connection, 
                     const gchar      *name,
                     gpointer          user_data)
{
  LunchboxApplication *self = user_data;
  GtkWindow *window;

  g_assert (LUNCHBOX_IS_APPLICATION (self));
  g_assert (G_IS_DBUS_CONNECTION (connection));

  window = gtk_application_get_active_window (GTK_APPLICATION (self));

  /**
   * We need to know if we already started the process or not
   * as the vanished_cb might get called the moment the watcher is registered
   * and we don't want to display an error when the qemu process is starting.
   */
  if (self->process_started)
    {
      lunchbox_window_set_status_view (LUNCHBOX_WINDOW(window), NULL, 
                                      _("Process Stopped"),
                                      _("The DBus name is no longer available"));
      if (self->proc)
      {
        g_subprocess_force_exit (self->proc);
        self->proc = NULL;
      }
    }
  else
    {
      lunchbox_window_set_loading_view (LUNCHBOX_WINDOW (window));
    }
}

static void
start_qemu (LunchboxApplication *self)
{
  GtkWindow *window;
  GSubprocess *proc;
  g_autoptr (GError) error = NULL;
  g_assert (LUNCHBOX_IS_APPLICATION (self));

  if (self->argv && g_strv_length (self->argv) > 0)
    {
      g_info ("Starting qemu subprocess");
      proc = g_subprocess_newv ((const gchar *const *)self->argv,
                                G_SUBPROCESS_FLAGS_STDERR_MERGE, &error);
      if (error != NULL)
        {
          window = gtk_application_get_active_window (GTK_APPLICATION (self));
          lunchbox_window_set_status_view (LUNCHBOX_WINDOW (window), NULL,
                                          _("Failed to start QEMU"),
                                          error->message);
          return;
        } 
      else
        {
          self->proc = proc;
        }
    }
  g_info ("Starting bus watcher");
  self->watcher_id = g_bus_watch_name (G_BUS_TYPE_SESSION, "org.qemu",
                                       G_BUS_NAME_WATCHER_FLAGS_NONE,
                                       on_name_appeared_cb,
                                       on_name_vanished_cb,
                                       self,
                                       g_object_unref);
}

LunchboxApplication *
lunchbox_application_new (const char        *application_id,
                          GApplicationFlags  flags)
{
  g_return_val_if_fail (application_id != NULL, NULL);

  return g_object_new (LUNCHBOX_TYPE_APPLICATION,
                       "application-id", application_id,
                       "flags", flags,
                       NULL);
}

static void
lunchbox_application_activate (GApplication *app)
{
  GtkWindow *window;
  GtkSettings *settings;

  LunchboxApplication *self = LUNCHBOX_APPLICATION (app);

  settings = gtk_settings_get_default ();
  g_object_set (settings, "gtk-application-prefer-dark-theme", TRUE, NULL);

  g_assert (LUNCHBOX_IS_APPLICATION (app));
  window = gtk_application_get_active_window (GTK_APPLICATION (app));

  if (window == NULL)
    {
      window = g_object_new (LUNCHBOX_TYPE_WINDOW,
                             "application", app,
                             "fullscreened", self->is_fullscreened,
                             NULL);

      lunchbox_window_set_decorated (LUNCHBOX_WINDOW(window),
                                     self->is_decorated);
      lunchbox_window_set_emulate_touch (LUNCHBOX_WINDOW (window),
                                         self->emulate_touch);
      lunchbox_window_set_auto_resize (LUNCHBOX_WINDOW (window),
                                       self->auto_resize);
      lunchbox_window_show_host_cursor (LUNCHBOX_WINDOW (window),
                                        self->show_host_cursor);

      g_signal_connect_swapped (window,
                                "close-request",
                                G_CALLBACK (g_application_quit),
                                app);
      start_qemu (self);
    }
  gtk_window_present (window);
}

static gboolean
lunchbox_application_local_command_line (GApplication   *app,
                                         gchar        ***args,
                                         int            *exit_status)
{
  int i = 1;
  gchar **argv;
  LunchboxApplication *self = LUNCHBOX_APPLICATION (app);

  argv = *args;
  /* Used later to start QEMU */
  while (argv[i])
    {
      if (g_strcmp0 (argv[i], "--fullscreen") != 0 &&
          g_strcmp0 (argv[i], "--undecorate") != 0 &&
          g_strcmp0 (argv[i], "--emulate-touch") != 0 &&
          g_strcmp0 (argv[i], "--show-host-cursor") != 0 &&
          g_strcmp0 (argv[i], "--auto-resize") != 0)
        self->argv = g_strsplit (argv[i], " ", -1);
      else if (g_strcmp0 (argv[i], "--fullscreen") == 0)
        self->is_fullscreened = TRUE;
      else if (g_strcmp0 (argv[i], "--undecorate") == 0)
        self->is_decorated = FALSE;
      else if (g_strcmp0 (argv[i], "--emulate-touch") == 0)
        self->emulate_touch = TRUE;
      else if (g_strcmp0 (argv[i], "--show-host-cursor") == 0)
        self->show_host_cursor = TRUE;
      else if (g_strcmp0 (argv[i], "--auto-resize") == 0)
        self->auto_resize = TRUE;
      i++;
    }

  return G_APPLICATION_CLASS (lunchbox_application_parent_class)->local_command_line (app, args, exit_status);
}

static int
lunchbox_application_command_line (GApplication            *app,
                                   GApplicationCommandLine *cmdline)
{
  g_application_activate (app);
  return 0;
}

static void
lunchbox_application_shutdown (GApplication *app)
{
  LunchboxApplication *self = LUNCHBOX_APPLICATION (app);

  if (self->proc)
    g_subprocess_force_exit (self->proc);

  G_APPLICATION_CLASS (lunchbox_application_parent_class)->shutdown (app);
}

static void
lunchbox_application_finalize (GObject *object)
{
  LunchboxApplication *self = LUNCHBOX_APPLICATION (object);

  if (self->watcher_id > 0)
    g_bus_unwatch_name (self->watcher_id);
  G_OBJECT_CLASS (lunchbox_application_parent_class)->finalize (object);
}

static void
lunchbox_application_dispose (GObject *object)
{
  LunchboxApplication *self = LUNCHBOX_APPLICATION (object);
  g_clear_pointer (&self->argv, g_strfreev);

  G_OBJECT_CLASS (lunchbox_application_parent_class)->dispose (object);
}

static void
lunchbox_application_class_init (LunchboxApplicationClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS(klass);
  GApplicationClass *app_class = G_APPLICATION_CLASS(klass);

  object_class->dispose   = lunchbox_application_dispose;
  object_class->finalize  = lunchbox_application_finalize;

  app_class->activate           = lunchbox_application_activate;
  app_class->shutdown           = lunchbox_application_shutdown;
  app_class->local_command_line = lunchbox_application_local_command_line;
  app_class->command_line       = lunchbox_application_command_line;
}

static void
lunchbox_application_quit_action (GSimpleAction *action,
                                  GVariant      *parameter,
                                  gpointer       user_data)
{
  LunchboxApplication *self = user_data;
  g_assert (LUNCHBOX_IS_APPLICATION (self));

  g_application_quit (G_APPLICATION (self));
}

static const GActionEntry app_actions[] = {
  { "quit", lunchbox_application_quit_action },
};

static GOptionEntry options[] =
{
    { "undecorate", 0, 0, G_OPTION_ARG_NONE, NULL,
      N_("Disable window decoration"), NULL },
    { "fullscreen", 0, 0, G_OPTION_ARG_NONE, NULL,
      N_("Start the window fullscreened"), NULL },
    { "emulate-touch", 0, 0, G_OPTION_ARG_NONE, NULL,
      N_("Emulate touch from mouse motion and click events"), NULL },
    { "show-host-cursor", 0, 0, G_OPTION_ARG_NONE, NULL,
      N_("Display host cursor"), NULL },
    { "auto-resize", 0, 0, G_OPTION_ARG_NONE, NULL,
      N_("Enable auto reconfiguring the guest OS's display size when resized"), NULL },
    { NULL }
};

static void
lunchbox_application_init (LunchboxApplication *self)
{
  self->process_started = FALSE;
  self->is_fullscreened = FALSE;
  self->is_decorated = TRUE;
  self->emulate_touch = FALSE;
  self->show_host_cursor = FALSE;
  self->auto_resize = FALSE;
  g_action_map_add_action_entries (G_ACTION_MAP (self),
                                   app_actions,
                                   G_N_ELEMENTS (app_actions),
                                   self);
  g_application_add_main_option_entries (G_APPLICATION (self), options);
  gtk_application_set_accels_for_action (GTK_APPLICATION (self),
                                         "app.quit",
                                         (const char *[]){ "<primary>q", NULL });
  gtk_application_set_accels_for_action (GTK_APPLICATION (self),
                                         "win.fullscreen",
                                         (const char *[]){ "<primary>F11", NULL });
  gtk_application_set_accels_for_action (GTK_APPLICATION (self),
                                         "vm.screenshot",
                                         (const char *[]){ "<primary><alt><shift>S", NULL });
}

